import inspect
import os
import shutil
import stat
import subprocess
import sys
import venv
from pathlib import Path
from typing import List

from git import GitCommandError, InvalidGitRepositoryError, Repo
from packaging.version import parse as parse_version
from snk_cli.config.config import SnkConfig
from snk_cli.workflow import Workflow

from .errors import (
    InvalidWorkflowError,
    InvalidWorkflowRepositoryError,
    WorkflowExistsError,
    WorkflowNotFoundError,
)


class Nest:
    """
    Initializes a Nest object.

    Args:
      snk_home (Path, optional): The path to the SNK home directory. Defaults to None.
      bin_dir (Path, optional): The path to the bin directory. Defaults to None.

    Side Effects:
      Creates the SNK home and bin directories if they do not exist.

    Examples:
      >>> nest = Nest()
    """

    def __init__(self, snk_home: Path = None, bin_dir: Path = None) -> None:
        """
        Initializes a Nest object.

        Args:
          snk_home (Path, optional): The path to the SNK home directory. Defaults to None.
          bin_dir (Path, optional): The path to the bin directory. Defaults to None.

        Side Effects:
          Creates the SNK home and bin directories if they do not exist.

        Examples:
          >>> nest = Nest()
        """
        self.python_interpreter_path = Path(
            sys.executable
        )  # needs to be the same python that has snk

        if not snk_home:
            home_path = self.python_interpreter_path.parent.parent
            if not os.access(home_path, os.W_OK):
                user_home_path = Path("~").expanduser()
                snk_home = user_home_path / ".local" / "snk"
            else:
                snk_home = home_path / "snk"

        if not bin_dir:
            bin_dir = self.python_interpreter_path.parent
            if not os.access(bin_dir, os.W_OK):
                user_home_path = Path("~").expanduser()
                bin_dir = user_home_path / ".local" / "bin"

        self.bin_dir = Path(bin_dir).absolute()
        self.snk_home = Path(snk_home).absolute()
        self.snk_workflows_dir = self.snk_home / "workflows"
        self.snk_venv_dir = self.snk_home / "venvs"
        self.snk_executable_dir = self.snk_home / "bin"

        # Create dirs
        self.snk_home.mkdir(parents=True, exist_ok=True)
        self.snk_workflows_dir.mkdir(parents=True, exist_ok=True)
        self.snk_executable_dir.mkdir(parents=True, exist_ok=True)
        self.bin_dir.mkdir(parents=True, exist_ok=True)

    def bin_dir_in_path(self) -> bool:
        path_dirs = os.environ["PATH"].split(os.pathsep)
        return str(self.bin_dir) in path_dirs

    def _format_repo_url(self, repo: str):
        """
        Checks that the given repo URL is valid.

        Args:
          repo (str): The URL of the repo.

        Raises:
          InvalidWorkflowRepositoryError: If the repo URL is not valid.

        Examples:
          >>> nest._format_repo_url("https://github.com/example/repo.git")
        """
        if not repo.endswith(".git"):
            repo += ".git"
        if not repo.startswith("http"):
            raise InvalidWorkflowRepositoryError("Repo url must start with http")
        return repo

    def install(
        self,
        workflow: str,
        editable=False,
        name=None,
        tag=None,
        commit=None,
        config: Path = None,
        snakefile: Path = None,
        force=False,
        additional_resources=[],
        conda: bool = None,
        snakemake_version=None,
        dependencies=[],
        isolate=False,
    ) -> Workflow:
        """
        Installs a Snakemake workflow as a CLI.

        Args:
          workflow (str): The URL of the repo or the path to the local workflow.
          editable (bool, optional): Whether to install the workflow in editable mode. Defaults to False.
          name (str, optional): The name of the workflow. Defaults to None.
          tag (str, optional): The tag of the workflow. Defaults to None.
          commit (str, optional): The commit SHA of the workflow. Defaults to None.
          config (Path, optional): The path to the snakemake config file. Defaults to None.
          snakefile (Path, optional): The path to the Snakefile. Defaults to None.
          force (bool, optional): Whether to force the installation. Defaults to False.
          additional_resources (list, optional): A list of resources additional to the resources folder to copy. Defaults to [].
          conda (bool, optional): Modify the snk config file to control conda use. If None, will not modify the config file. Defaults to None.
          snakemake_version (str, optional): The version of Snakemake to install in the virtual environment. Defaults to None.
          dependencies (list, optional): A list of dependencies to install. Defaults to [].
        Returns:
          Workflow: The installed workflow.

        Examples:
          >>> nest.install(
          ...     "https://github.com/example/repo.git", name="example", tag="v1.0.0"
          ... )
          >>> nest.install(
          ...     "https://github.com/example/repo.git", name="example", commit="0123456"
          ... )
        """

        def handle_force_installation(name: str):
            try:
                self.uninstall(name=name, force=True)
            except WorkflowNotFoundError:
                pass

        workflow = str(workflow)  # ensure it is a string
        try:
            workflow = self._format_repo_url(workflow)
            if not name:
                name = self._get_name_from_git_url(workflow)
            if not force:
                self._check_workflow_name_available(name)
            else:
                handle_force_installation(name)
            workflow_path = self.download(workflow, name, tag_name=tag, commit=commit)
        except WorkflowNotFoundError as e:
            to_remove = self.get_paths_to_delete(name)
            self.delete_paths(to_remove)
            raise e
        except InvalidWorkflowRepositoryError:
            workflow_local_path = Path(workflow).resolve()
            if workflow_local_path.is_file():
                raise InvalidWorkflowError(
                    f"When installing a local workflow, the path must be a directory. Found: {workflow_local_path}"
                )
            if (
                self.snk_workflows_dir.resolve().is_relative_to(workflow_local_path)
                and not editable
            ):
                raise InvalidWorkflowError(
                    f"The workflow directory contains SNK_HOME!\nWORKFLOW: {workflow_local_path}\nSNK_HOME: {self.snk_workflows_dir.resolve()}.\n\nTry installing the workflow with --editable."
                )
            if not name:
                name = workflow_local_path.name
            if not force:
                self._check_workflow_name_available(name)
            else:
                handle_force_installation(name)
            workflow_path = self.local(workflow_local_path, name, editable)
        try:
            self.validate_Snakemake_repo(workflow_path)
            # update non standard files
            if config:
                config_path = workflow_path / config
                if not config_path.exists():
                    raise FileNotFoundError(f"Config file not found at {config_path}")
                self.modify_snk_config(workflow_path, configfile=config_path)
            if snakefile:
                snakefile_path = workflow_path / snakefile
                if not snakefile_path.exists():
                    raise FileNotFoundError(f"Snakefile not found at {snakefile_path}")
                self.modify_snk_config(workflow_path, snakefile=snakefile_path)
            # set the version of the workflow
            if editable:
                version = "editable"
            elif tag:
                version = tag
            elif commit:
                version = commit
            else:
                try:
                    repo = Repo(workflow_path)
                    sha = repo.head.object.hexsha
                    version = repo.git.rev_parse(sha, short=8)
                except Exception:
                    version = None
            self.modify_snk_config(workflow_path, version=version)
            # check if we need to install snakemake in a virtual environment
            snakemake_version_to_install_in_venv = None
            snakemake_min_version = self.check_for_snakemake_min_version(workflow_path, snakefile)
            if snakemake_version is not None:
                snakemake_version_to_install_in_venv = snakemake_version
                if parse_version(self._current_snakemake_version) < parse_version(
                    snakemake_min_version
                ):
                    # The current version of Snakemake is less than the minimum version required by the workflow
                    snakemake_version_to_install_in_venv = f">={snakemake_min_version}"
            min_snk_cli_version = self.check_for_snk_cli_min_version(workflow_path)
            if min_snk_cli_version is not None:
                if parse_version(self._current_snk_cli_version) < parse_version(
                    min_snk_cli_version
                ):
                    # The current version of Snakemake is less than the minimum version required by the workflow
                    dependencies.append(f"snk_cli>={min_snk_cli_version}")
            if snakemake_version_to_install_in_venv is not None or dependencies:
                isolate = True
            if isolate:
                venv_path = self.create_virtual_environment(name)
                self._install_snk_cli_in_venv(
                    venv_path,
                    snakemake_version=snakemake_version_to_install_in_venv,
                    dependencies=dependencies,
                )
                python_interpreter_path = venv_path / "bin" / "python"
            else:
                python_interpreter_path = self.python_interpreter_path
            # create the workflow executable
            workflow_executable_path = self.create_executable(
                workflow_path, name, python_interpreter_path=python_interpreter_path
            )
            self.link_workflow_executable_to_bin(workflow_executable_path)
            if additional_resources:
                self.additional_resources(workflow_path, additional_resources)
            if conda is not None:
                self.modify_snk_config(workflow_path, conda=conda)
            self._confirm_installation(name)
        except Exception as e:
            # remove any half completed steps
            to_remove = self.get_paths_to_delete(name)
            self.delete_paths(to_remove)
            raise e
        return Workflow(workflow_path)

    def modify_snk_config(self, workflow_path: Path, **kwargs):
        """
        Modify the snk config file.

        Args:
          workflow_path (Path): The path to the workflow directory.
          **kwargs: Additional keyword arguments to modify the snk config file.

        Examples:
          >>> nest.modify_snk_config(Path("/path/to/workflow"), logo=example)
        """
        snk_config = SnkConfig.from_workflow_dir(workflow_path, create_if_not_exists=True)
        modified = False
        for key, value in kwargs.items():
            if getattr(snk_config, key) != value:
                modified = True
                setattr(snk_config, key, value)
        if modified:
            snk_config.save()

    def additional_resources(self, workflow_path: Path, resources: List[Path]):
        """
        Modify the snk config file so that resources will be copied at runtime.

        Args:
          workflow_path (Path): The path to the workflow directory.
          resources (List[Path]): A list of additional resources to copy.

        Examples:
          >>> nest.additional_resources(
          ...     Path("/path/to/workflow"),
          ...     [Path("/path/to/resource1"), Path("/path/to/resource2")],
          ... )
        """
        # validate_resources(resources)
        snk_config = SnkConfig.from_workflow_dir(workflow_path, create_if_not_exists=True)
        snk_config.add_resources(resources, workflow_path)
        snk_config.save()

    def copy_nonstandard_config(self, workflow_dir: Path, config_path: Path):
        """
        Copy a nonstandard config file to the workflow directory.

        Args:
          workflow_dir (Path): The path to the workflow directory.
          config_path (Path): The path to the config file.

        Examples:
          >>> nest.copy_nonstandard_config(
          ...     Path("/path/to/workflow"), Path("/path/to/config.yaml")
          ... )
        """
        config_dir = workflow_dir / "config"
        config_dir.mkdir(exist_ok=True)
        shutil.copyfile(workflow_dir / config_path, config_dir / "config.yaml")

    def get_paths_to_delete(self, workflow_name: str) -> List[Path]:
        """
        Get the paths to delete when uninstalling a workflow.

        Args:
          workflow_name (str): The name of the workflow.

        Returns:
          List[Path]: A list of paths to delete.

        Examples:
          >>> nest.get_paths_to_delete("example")
          [Path('/path/to/workflows/example'), Path('/path/to/bin/example')]
        """
        to_delete = []

        # remove repo
        workflow_dir = self.snk_workflows_dir / workflow_name
        if workflow_dir.exists() and workflow_dir.is_dir():
            to_delete.append(workflow_dir)
        elif workflow_dir.is_symlink():
            # editable
            to_delete.append(workflow_dir)

        workflow_executable = self.snk_executable_dir / workflow_name
        if workflow_executable.exists():
            to_delete.append(workflow_executable)

        # remove venv
        venv_path = self.snk_venv_dir / workflow_name
        if venv_path.exists():
            to_delete.append(venv_path)

        # remove link
        workflow_symlink_executable = self.bin_dir / workflow_name
        if workflow_symlink_executable.is_symlink():
            if str(os.readlink(workflow_symlink_executable)) == str(workflow_executable):
                to_delete.append(workflow_symlink_executable)

        return to_delete

    def delete_paths(self, files: List[Path]):
        """
        Delete the given paths.

        Args:
          files (List[Path]): A list of paths to delete.

        Side Effects:
          Deletes the given paths.

        Examples:
          >>> nest.delete_paths(
          ...     [Path("/path/to/workflows/example"), Path("/path/to/bin/example")]
          ... )
        """
        # check that the files are in self.snk_workflows_dir
        # i.e. if it is a symlink read the link and check
        for path in files:
            if path.is_symlink():
                print("Unlinking:", path)
                path.unlink()
            elif path.is_file():
                print("Deleting:", path)
                assert str(self.snk_home) in str(path), "Cannot delete files outside of SNK_HOME"
                path.unlink()
            elif path.is_dir():
                print("Deleting:", path)
                assert str(self.snk_home) in str(path), "Cannot delete folders outside of SNK_HOME"
                shutil.rmtree(path)
            else:
                raise TypeError("Invalid file type")

    def uninstall(self, name: str, force: bool = False) -> bool:
        """
        Uninstalls a workflow.

        Args:
          name (str): The name of the workflow.
          force (bool, optional): Whether to force the uninstallation. Defaults to False.

        Returns:
          bool: Whether the uninstallation was successful.

        Examples:
          >>> nest.uninstall("example")
          True
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string. Found: {name}")
        to_remove = self.get_paths_to_delete(name)
        if not to_remove:
            raise WorkflowNotFoundError(f"Workflow '{name}' not found")
        if force:
            proceed = True
        else:
            print(f"Uninstalling {name}")
            print("  Would remove:")
            for p in to_remove:
                print(f"    {p}{'/*' if p.is_dir() else ''}")
            ans = input("Proceed (Y/n)? ")
            proceed = ans.lower() in ["y", "yes", ""]
        if not proceed:
            return False
        self.delete_paths(to_remove)
        return True

    def _check_workflow_name_available(self, name: str):
        if not name:
            return None
        if name in os.listdir(self.snk_workflows_dir):
            raise WorkflowExistsError(
                f"Workflow '{name}' already exists in SNK_HOME ({self.snk_workflows_dir})"
            )

        if name in os.listdir(self.bin_dir):
            # check if orfan symlink
            def is_orfan_symlink(name):
                return (self.bin_dir / name).is_symlink() and str(self.snk_home) in os.readlink(
                    self.bin_dir / name
                )

            if is_orfan_symlink(name):
                self.delete_paths([self.bin_dir / name])
            else:
                raise WorkflowExistsError(
                    f"File '{name}' already exists in SNK_BIN ({self.bin_dir})"
                )

    def _confirm_installation(self, name: str):
        """
        Confirms that the installation was successful.

        Args:
          name (str): The name of the workflow.

        Examples:
          >>> nest._confirm_installation("example")
        """
        workflow_dir = self.snk_workflows_dir / name
        assert workflow_dir.exists()
        workflows = [p.name.split(".")[0] for p in self.bin_dir.glob("*")]
        assert name in workflows

    def _get_name_from_git_url(self, git_url: str):
        """
        Gets the name of the workflow from the git URL.

        Args:
          git_url (str): The URL of the git repository.

        Returns:
          str: The name of the workflow.
        """
        return git_url.split("/")[-1].split(".")[0]

    @property
    def workflows(self):
        return [
            Workflow(workflow_dir.absolute()) for workflow_dir in self.snk_workflows_dir.glob("*")
        ]

    def download(self, repo_url: str, name: str, tag_name: str = None, commit: str = None) -> Path:
        """
        Clone a workflow from a git repository.

        Args:
          repo_url (str): The URL of the repo.
          name (str): The name of the workflow.
          tag_name (str, optional): The tag of the workflow. Defaults to None.
          commit (str, optional): The commit SHA of the workflow. Defaults to None.

        Returns:
          Path: The path to the cloned workflow.

        Examples:
          >>> nest.download(
          ...     "https://github.com/example/repo.git", "example", tag_name="v1.0.0"
          ... )
        """
        location = self.snk_workflows_dir / name
        options = []
        if not commit:
            options.append("--depth 1")
        if tag_name:
            options.append("--single-branch")
            options.append(f"--branch {tag_name}")
        try:
            repo = Repo.clone_from(repo_url, location, multi_options=options)
            if commit:
                repo.git.checkout(commit)
            else:
                repo.git.checkout(tag_name)
        except GitCommandError as e:
            if "destination path" in e.stderr:
                raise WorkflowExistsError(
                    f"Workflow '{name}' already exists in {self.snk_workflows_dir}"
                )
            elif f"Remote branch {tag_name}" in e.stderr:
                did_you_mean = ""
                if len(tag_name) < 6:
                    did_you_mean = f". Did you mean 'v{tag_name}'?"
                raise WorkflowNotFoundError(f"Workflow tag '{tag_name}' not found{did_you_mean}")
            elif f"pathspec '{commit}' did not match" in e.stderr:
                if tag_name:
                    raise WorkflowNotFoundError(
                        f"Workflow commit '{commit}' not found on branch {tag_name}"
                    )
                else:
                    raise WorkflowNotFoundError(f"Workflow commit '{commit}' not found")
            elif "not found" in e.stderr:
                raise WorkflowNotFoundError(f"Workflow repository '{repo_url}' not found")
            raise e
        return location

    def local(self, path: Path, name: str, editable=False) -> Path:
        """
        Install a local workflow.

        Args:
          path (Path): The path to the local workflow.
          name (str): The name of the workflow.
          editable (bool, optional): Whether to install the workflow in editable mode. Defaults to False.

        Returns:
          Path: The path to the installed workflow.

        Examples:
          >>> nest.local(Path("/path/to/workflow"), "example")
        """
        location = self.snk_workflows_dir / name
        if editable:
            os.symlink(path.absolute(), location, target_is_directory=True)
            return location
        shutil.copytree(path, location)
        try:
            Repo(location)
        except InvalidGitRepositoryError:
            Repo.init(location, mkdir=False)
        return location

    def create_virtual_environment(self, name: str) -> Path:
        """
        Create a virtual environment for the workflow.

        Args:
          name (str): The name of the virtual environment.
          snakemake_version (str, optional): The version of Snakemake to install in the virtual environment. Defaults to None.

        Returns:
          Path: The path to the virtual environment.

        Examples:
          >>> nest.create_virtual_environment("example")
        """
        venv_dir = self.snk_home / "venvs"
        venv_dir.mkdir(exist_ok=True)
        venv_path = venv_dir / name
        try:
            venv.create(venv_path, with_pip=True, symlinks=True)
        except FileExistsError:
            raise FileExistsError(
                f"The venv {venv_path} already exists. Please choose a different location or name. Alternatively, use the --force flag to overwrite the existing venv."
            )
        return venv_path

    def _install_snk_cli_in_venv(self, venv_path: Path, snakemake_version=None, dependencies=[]):
        """
        Install snk_cli in the virtual environment.

        Args:
          venv_path (Path): The path to the virtual environment.
          snakemake_version (str, optional): The version of Snakemake to install in the virtual
          environment. Defaults to "7.32.4".

        Examples:
          >>> nest._install_snk_cli_in_venv(Path("/path/to/venv"), snakemake_version="7.32.4")
        """
        if not venv_path.exists():
            raise FileNotFoundError(f"Virtual environment not found at {venv_path}")
        if sys.platform.startswith("win"):
            pip_path = venv_path / "Scripts" / "pip.exe"
        else:
            pip_path = venv_path / "bin" / "pip"
        if not pip_path.exists():
            raise FileNotFoundError(f"pip not found at {pip_path}")
        # check if snakemake version starts with >= or <=
        if snakemake_version:
            # check if snakemake version starts with >, <, =, ~, ^,
            if snakemake_version[0] in [">", "<", "="]:
                snakemake_version = f"snakemake{snakemake_version}"
            else:
                snakemake_version = f"snakemake=={snakemake_version}"
        else:
            snakemake_version = "snakemake"
        try:
            snk_cli_in_deps = len([dep for dep in dependencies if "snk_cli" in dep]) > 0
            if not snk_cli_in_deps:
                dependencies.append("snk_cli")
            subprocess.run(
                [pip_path, "install", snakemake_version, "setuptools"] + dependencies,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise Exception(f"Failed to install snk_cli in virtual environment. Error: {e}")

    def create_executable(
        self, workflow_path: Path, name: str, python_interpreter_path=None
    ) -> Path:
        if not python_interpreter_path:
            python_interpreter_path = self.python_interpreter_path
        template = inspect.cleandoc(
            f"""
            #!/bin/sh
            '''exec' "{python_interpreter_path}" "$0" "$@"
            ' '''
            # -*- coding: utf-8 -*-
            import re
            import sys
            from pathlib import Path
            from snk_cli import CLI

            def create_cli(p):
                workflow_dir_path = Path(p)
                cli = CLI(workflow_dir_path)
                cli()

            if __name__ == "__main__":
                sys.argv[0] = re.sub(r'(-script\\.pyw|\\.exe)?$', '', sys.argv[0])
                sys.exit(create_cli("{workflow_path}"))
                
        """
        )

        if sys.platform.startswith("win"):
            name += ".exe"

        workflow_executable = self.snk_executable_dir / name

        with open(workflow_executable, "w") as f:
            f.write(template)

        workflow_executable.chmod(workflow_executable.stat().st_mode | stat.S_IEXEC)

        return workflow_executable

    def link_workflow_executable_to_bin(self, workflow_executable_path: Path):
        """
        Links a workflow executable to the bin directory.

        Args:
          workflow_executable_path (Path): The path to the workflow executable.

        Returns:
          Path: The path to the linked workflow executable.

        Examples:
          >>> nest.link_workflow_executable_to_bin(Path("/path/to/workflow_executable"))
        """
        name = workflow_executable_path.name
        if (self.bin_dir / name).is_symlink() and os.readlink(self.bin_dir / name) == str(
            workflow_executable_path
        ):
            # skip if it's already there
            return self.bin_dir / name
        try:
            os.symlink(workflow_executable_path.absolute(), self.bin_dir / name)
        except FileExistsError:
            raise WorkflowExistsError(f"File '{name}' already exists in SNK_BIN ({self.bin_dir})")
        return self.bin_dir / name

    def check_for_snakemake_min_version(self, workflow_path: Path, snakefile: Path = None):
        """
        Check if the workflow has a minimum version of Snakemake.

        Args:
          workflow_path (Path): The path to the workflow directory.

        Returns:
          str: The minimum version of Snakemake.

        Examples:
          >>> nest.check_for_snakemake_min_version(Path("/path/to/workflow"))
        """
        import re

        min_version = "0.0.0"
        if snakefile:
            snakefile = workflow_path / snakefile
        else:
            for snakefile in workflow_path.glob("**/Snakefile"):
                if snakefile.exists():
                    break
            else:
                return min_version
        with open(snakefile, "r") as f:
            for line in f:
                match = re.search(r"min_version\((.*)\)", line)
                if match:
                    min_version = match.group(1).strip().strip('"').strip("'")
                    break
        return min_version

    @property
    def _current_snakemake_version(self):
        """
        Get the current version of Snakemake.

        Returns:
          str: The current version of Snakemake.

        Examples:
          >>> nest.get_current_current_snakemake_version()
        """
        from snakemake.common import __version__

        return __version__

    def check_for_snk_cli_min_version(self, workflow_path: Path):
        """
        Check if the workflow has a minimum version of snk_cli.

        Args:
          workflow_path (Path): The path to the workflow directory.

        Returns:
          str: The minimum version of snk_cli.

        Examples:
          >>> nest.check_for_snk_cli_min_version(Path("/path/to/workflow"))
        """
        snk_config = SnkConfig.from_workflow_dir(workflow_path, create_if_not_exists=True)
        try:
            return snk_config.min_snk_cli_version
        except AttributeError:
            return None
    
    @property
    def _current_snk_cli_version(self):
        """
        Get the current version of snk_cli.

        Returns:
          str: The current version of snk_cli.

        Examples:
          >>> nest.get_current_current_snk_cli_version()
        """
        from snk_cli.__about__ import __version__

        return __version__

    def validate_Snakemake_repo(self, repo: Repo):
        """
        Validates a Snakemake repository.

        Args:
          repo (Repo): The Snakemake repository.

        Returns:
          bool: True if the repository is valid, False otherwise.

        Examples:
          >>> nest.validate_Snakemake_repo(Repo("/path/to/repo"))
        """
        pass



class NestError(Exception):
    """
    Base class for all workflow exceptions.
    """


class WorkflowExistsError(NestError):
    """
    Thrown if the given workflow is already installed.
    """


class WorkflowNotFoundError(NestError):
    """
    Thrown if the given workflow cannot be found.
    """


class InvalidWorkflowRepositoryError(NestError):
    """
    Thrown if the given repository appears to have an invalid format.
    """


class InvalidWorkflowError(NestError):
    """
    Thrown if the given workflow appears to have an invalid format.
    """



from pathlib import Path
from typing import Tuple

import pytest
import yaml
from snk_cli.config import SnkConfig

from snk import Nest

from .utils import CLIRunner


@pytest.fixture()
def bin_dir(tmp_path_factory: Path):
    return tmp_path_factory.mktemp("bin")


@pytest.fixture()
def snk_home(tmp_path_factory: Path):
    return tmp_path_factory.mktemp("snk")


@pytest.fixture()
def nest(snk_home, bin_dir):
    return Nest(snk_home, bin_dir)


@pytest.fixture()
def example_config():
    return Path("tests/data/config.yaml")


@pytest.fixture(scope="session")
def basic_runner(tmp_path_factory):
    nest = Nest(tmp_path_factory.mktemp("snk"), tmp_path_factory.mktemp("bin"))
    basic_workflow = nest.install("https://github.com/Wytamma/snk-basic-pipeline.git")
    expected = nest.snk_workflows_dir / "snk-basic-pipeline"
    assert expected.exists()
    print(basic_workflow.executable)
    runner = CLIRunner([basic_workflow.executable])
    return runner


@pytest.fixture(scope="module")
def local_workflow(tmp_path_factory):
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    print(nest.bin_dir)
    print(nest.snk_home)
    local_workflow = nest.install("tests/data/workflow", editable=True)
    expected = nest.snk_workflows_dir / "workflow"
    assert expected.exists()
    return local_workflow


@pytest.fixture(scope="session")
def local_runner(tmp_path_factory):
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    local_workflow = nest.install("tests/data/workflow", editable=True)
    runner = CLIRunner([local_workflow.executable])
    return runner


class Request:
    param: Tuple[dict, SnkConfig]


@pytest.fixture
def dynamic_runner(tmp_path_factory, request: Request) -> CLIRunner:
    """Create a CLI Runner from a SNK and config file"""
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    config, snk, snakefile_text = request.param
    path = Path(tmp_path_factory.mktemp("workflow"))
    snk_path = path / "snk.yaml"
    config_path = path / "config.yaml"
    with open(config_path, "w") as f:
        yaml.dump(config, f)
    snk.to_yaml(snk_path)
    snakefile_path = path / "Snakefile"
    snakefile_path.write_text(snakefile_text)
    workflow = nest.install(path, editable=True)
    runner = CLIRunner([workflow.executable])
    return runner


@pytest.fixture(scope="session")
def print_config_runner(tmp_path_factory):
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    local_workflow = nest.install("tests/data/print_config", editable=True)
    runner = CLIRunner([local_workflow.executable])
    return runner



import os  # noqa: F401
import platform  # noqa: F401
import subprocess
from pathlib import Path
from unittest.mock import call, patch

import pytest
import typer  # noqa: F401
from snk_cli.workflow import Workflow
from typer.testing import CliRunner

from snk.main import app

runner = CliRunner()


@pytest.fixture
def mock_platform_system():
    with patch('platform.system') as mock:
        yield mock


@pytest.fixture
def mock_os_startfile():
    # Only patch os.startfile if the platform is Windows
    if hasattr(os, 'startfile'):
        with patch('os.startfile') as mock:
            yield mock
    else:
        yield None


@pytest.fixture
def mock_subprocess_call():
    with patch('subprocess.call') as mock:
        yield mock


@pytest.mark.skipif(platform.system() != 'Windows', reason="Requires Windows")
def test_open_text_editor_windows(
    mock_platform_system, mock_os_startfile, local_workflow: Workflow
):
    mock_platform_system.return_value = 'Windows'
    file_path = local_workflow.path / "snk.yaml"
    
    snk_home = local_workflow.path.parent.parent
    bin_dir = local_workflow.path.parent.parent.parent / "bin"
    res = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "edit", "workflow"])
    
    assert res.exit_code == 0, res.stderr
    if mock_os_startfile:
        mock_os_startfile.assert_called_once_with(file_path)


def test_open_text_editor_mac(
    mock_platform_system, mock_subprocess_call, local_workflow: Workflow
):
    mock_platform_system.return_value = 'Darwin'
    file_path = local_workflow.path / "snk.yaml"

    snk_home = local_workflow.path.parent.parent
    bin_dir = local_workflow.path.parent.parent.parent / "bin"
    res = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "edit", "workflow"])
    
    assert res.exit_code == 0, res.stderr
    mock_subprocess_call.assert_called_once_with(('open', file_path))


def test_open_text_editor_linux(
    mock_platform_system, mock_subprocess_call, local_workflow: Workflow
):
    mock_platform_system.return_value = 'Linux'
    file_path = local_workflow.path / "snk.yaml"

    with patch('subprocess.call') as mock_call:
        mock_call.side_effect = [1, 1, 0, 0]  # Mocking 'which' command results: nano not found, vim not found, vi found

        snk_home = local_workflow.path.parent.parent
        bin_dir = local_workflow.path.parent.parent.parent / "bin"
        res = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "edit", "workflow"])
        assert res.exit_code == 0, res.stderr
        mock_call.assert_has_calls([
            call(['which', 'nano'], stdout=subprocess.PIPE, stderr=subprocess.PIPE),
            call(['which', 'vim'], stdout=subprocess.PIPE, stderr=subprocess.PIPE),
            call(['which', 'vi'], stdout=subprocess.PIPE, stderr=subprocess.PIPE),
            call(['vi', file_path])
        ])


def test_open_text_editor_no_editor_found(
    mock_platform_system, mock_subprocess_call, local_workflow: Workflow
):
    mock_platform_system.return_value = 'Linux'

    with patch('subprocess.call') as mock_call:
        mock_call.side_effect = [1, 1, 1]  # Mocking 'which' command results: none of the editors found

        with patch('typer.secho') as mock_print:
            snk_home = local_workflow.path.parent.parent
            bin_dir = local_workflow.path.parent.parent.parent / "bin"
            res = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "edit", "workflow"])
            assert res.exit_code == 1, res.stderr
            mock_print.assert_called_once_with(
                "No suitable text editor found. Please install nano or vim.", fg='red', err=True
            )



def open_text_editor(file_path):
    """
    Opens the system's default text editor to edit the specified file.

    Parameters:
    file_path (str): The path to the file to be edited.
    """
    import os
    import platform
    import subprocess

    if platform.system() == "Windows":
        os.startfile(file_path)
    elif platform.system() == "Darwin":  # macOS
        subprocess.call(("open", file_path))
    else:  # Linux and other Unix-like systems
        editors = ["nano", "vim", "vi"]
        editor = None
        for e in editors:
            if (
                subprocess.call(
                    ["which", e], stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
                == 0
            ):
                editor = e
                break
        if editor:
            subprocess.call([editor, file_path])
        else:
            raise Exception(
                "No suitable text editor found. Please install nano or vim."
            )


# SPDX-FileCopyrightText: 2022-present wytamma <wytamma.wirth@me.com>
#
# SPDX-License-Identifier: MIT
from pathlib import Path

from snk_cli import CLI, validate_config  # noqa: F401

from .nest import Nest  # noqa: F401


def create_cli(p):
    """
    This is the interface to create the dynamic CLI.
    Changing this function name will break backwards compatibility.

    Args:
      p (str): The path to the workflow directory.

    Returns:
      None

    Examples:
      >>> create_cli("/path/to/workflow")
      ... # CLI is created and executed
    """
    workflow_dir_path = Path(p)
    cli = CLI(workflow_dir_path)
    cli()



from pathlib import Path

from snk_cli.workflow import Workflow
from typer.testing import CliRunner

from snk.main import app

runner = CliRunner()


def test_snk_help():
    result = runner.invoke(app, ["-h"])
    assert result.exit_code == 0
    assert "snk" in result.stdout
    assert "Usage" in result.stdout


def test_snk_install(snk_home: Path, bin_dir: Path):
    result = runner.invoke(
        app, ["--home", snk_home, "--bin", bin_dir, "install", "tests/data/workflow"]
    )
    assert result.exit_code == 0
    assert "Successfully installed" in result.stdout
    assert (snk_home / "workflows" / "workflow").exists()
    assert (bin_dir / "workflow").is_symlink()


def test_snk_install_no_conda(snk_home: Path, bin_dir: Path):
    result = runner.invoke(
        app,
        [
            "--home",
            snk_home,
            "--bin",
            bin_dir,
            "install",
            "tests/data/workflow",
            "--no-conda",
        ],
    )
    assert result.exit_code == 0
    assert "Successfully installed" in result.stdout
    snk_config = snk_home / "workflows" / "workflow" / "snk.yaml"
    assert snk_config.exists()
    # read the snk.yaml file as dict
    import yaml

    with open(snk_config) as f:
        snk_config_dict = yaml.safe_load(f)
    assert snk_config_dict["conda"] is False


def test_snk_list(local_workflow: Workflow):
    snk_home = local_workflow.path.parent.parent
    bin_dir = local_workflow.path.parent.parent.parent / "bin"
    result = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "list"])
    assert result.exit_code == 0
    assert "workflow" in result.stdout
    assert "editable" in result.stdout

def test_config_path_cli(local_workflow: Workflow):
    snk_home = local_workflow.path.parent.parent
    bin_dir = local_workflow.path.parent.parent.parent / "bin"
    result = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "edit", "--path", "workflow"])
    assert result.exit_code == 0
    assert "/workflows/workflow/snk.yaml" in result.stdout  

def test_snk_uninstall(local_workflow: Workflow):
    snk_home = local_workflow.path.parent.parent
    bin_dir = local_workflow.path.parent.parent.parent / "bin"
    result = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "uninstall", "workflow"])
    assert "(Y/n)" in result.stdout
    result = runner.invoke(
        app, ["--home", snk_home, "--bin", bin_dir, "uninstall", "workflow", "--force"]
    )
    assert result.exit_code == 0
    assert "Successfully uninstalled" in result.stdout

def test_snk_create(local_workflow: Workflow, tmp_path: Path):
    result = runner.invoke(app, ["create", str(tmp_path), "--force"])
    print(result.stdout)
    assert result.exit_code == 0
    assert (tmp_path / "snk.yaml").exists()
    assert (tmp_path / "Snakefile").exists()

def test_import_create_cli(capsys):
    from snk import create_cli

    assert callable(create_cli)
    assert create_cli.__module__ == "snk"
    assert create_cli.__name__ == "create_cli"

    import sys

    sys.argv = ["cli", "-h"]
    try:
        create_cli("tests/data/workflow")
    except SystemExit:
        pass
    captured = capsys.readouterr()
    assert "Usage" in captured.out




from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from snk import Nest


@pytest.mark.parametrize(
    "args",
    [
        {"snakemake_version": None, "dependencies": []},
        {"snakemake_version": "7.0.0", "dependencies": []},
        {"snakemake_version": None, "dependencies": ["pandas", "numpy"]},
        {"snakemake_version": "7.0.0", "dependencies": ["pandas", "numpy"]},
    ],
)
def test_install_snk_cli_in_venv(nest: Nest, args):
    # Define the virtual environment path
    venv_path = Path("/fake/path/to/venv")

    # Mock `exists` method for both the venv_path and the pip_path
    with patch.object(Path, "exists", MagicMock(return_value=True)) as mock_exists:
        with patch("sys.platform", "linux"):
            # We assume a Linux platform for this example.
            # Adjust if different OS handling is required
            pip_path = venv_path / "bin" / "pip"
            with patch.object(Path, "__truediv__", return_value=pip_path):
                with patch("subprocess.run") as mock_run:
                    # Assume subprocess.run does not raise an exception
                    mock_run.return_value = None

                    # Call the method under test
                    nest._install_snk_cli_in_venv(venv_path, **args)

                    # Check that subprocess.run was called with the correct arguments
                    if args["snakemake_version"] is None:
                        snakemake = "snakemake"
                    else:
                        snakemake = f"snakemake=={args['snakemake_version']}"
                    mock_run.assert_called_once_with(
                        [pip_path, "install", snakemake, "setuptools"]
                        + args["dependencies"],
                        check=True,
                    )
                    # Verify that the Path.exists method was called on both venv_path and pip_path
                    mock_exists.assert_any_call()



from pathlib import Path
from types import SimpleNamespace
from typing import List, Optional

import typer
from rich.progress import Progress, SpinnerColumn, TextColumn
from snk_cli.config import SnkConfig

from .__about__ import __version__
from .errors import WorkflowExistsError, WorkflowNotFoundError
from .nest import Nest
from .utils import open_text_editor

app = typer.Typer()

SNK_HOME = None
SNK_BIN = None


def _print_snk_version(self, ctx: typer.Context, value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


# fmt: off
@app.callback(context_settings={"help_option_names": ["-h", "--help"]})
def callback(
    ctx: typer.Context,
    home: Optional[Path] = typer.Option(
            None, 
            envvar="SNK_HOME", 
            dir_okay=True, 
            file_okay=False, 
            exists=True, 
            help="Overrides default snk location. Workflows will be installed in $SNK_HOME/workflows."
        ),
    bin: Optional[Path] = typer.Option(
            None, 
            envvar="SNK_BIN", 
            dir_okay=True, 
            file_okay=False, 
            exists=True, 
            help="Overrides location of workflow installations. Workflows are symlinked here."
        ),
    version: Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        help="Show the version and exit.",
        is_eager=True,
        callback=_print_snk_version,
        show_default=False,
        ),
    ):
    # suppress python warning 
    ctx.obj = SimpleNamespace(snk_home = home, snk_bin = bin)
# fmt: on

callback.__doc__ = "\b"
callback.__doc__ += r"""        _            _             _        
       / /\         /\ \     _    /\_\      
      / /  \       /  \ \   /\_\ / / /  _   
     / / /\ \__   / /\ \ \_/ / // / /  /\_\ 
    / / /\ \___\ / / /\ \___/ // / /__/ / / 
    \ \ \ \/___// / /  \/____// /\_____/ /  
     \ \ \     / / /    / / // /\_______/   
 _    \ \ \   / / /    / / // / /\ \ \      
/_/\__/ / /  / / /    / / // / /  \ \ \     
\ \/___/ /  / / /    / / // / /    \ \ \    
 \_____\/   \/_/     \/_/ \/_/      \_\_\   
"""
callback.__doc__ += f"\b\n\nA Snakemake Workflow Management System ({__version__})"


@app.command()
def install(
    ctx: typer.Context,
    workflow: str = typer.Argument(
        ..., help="Path, URL or Github name (user/repo) of the workflow to install."
    ),
    name: Optional[str] = typer.Option(
        None,
        "--name",
        "-n",
        help="Rename the workflow (this name will be used to call the CLI.)",
    ),
    tag: Optional[str] = typer.Option(
        None,
        "--tag",
        "-t",
        help="Tag (version) of the workflow to install. Can specify a branch name, or tag. If None the latest commit will be installed.",
    ),
    commit: Optional[str] = typer.Option(
        None,
        "--commit",
        "-c",
        help="Commit (SHA) of the workflow to install. If None the latest commit will be installed.",
    ),
    isolate: Optional[bool] = typer.Option(
        False,
        "--isolate",
        "-i",
        help="Install the workflow in a isolated environment.",
    ),
    snakemake_version: Optional[str] = typer.Option(
        None,
        "--snakemake",
        "-s",
        help="Snakemake version to install with the isolated workflow. Default is the latest version.",
    ),
    dependencies: Optional[List[str]] = typer.Option(
        [],
        "--dependency",
        "-d",
        help="Additional pip dependencies to install with the workflow.",
    ),
    config: Optional[Path] = typer.Option(None, help="Specify a non-standard config location."),
    snakefile: Optional[Path] = typer.Option(
        None, help="Specify a non-standard Snakefile location."
    ),
    resource: Optional[List[Path]] = typer.Option(
        [],
        help="Specify resources additional to the resources folder required by the workflow (copied to working dir at runtime).",
    ),
    no_conda: bool = typer.Option(
        False,
        "--no-conda",
        help="Do not use conda environments by default.",
    ),
    force: Optional[bool] = typer.Option(
        False, "--force", "-f", help="Force install (overwrites existing installs)."
    ),
    editable: Optional[bool] = typer.Option(
        False,
        "--editable",
        "-e",
        help="Whether to install the workflow in editable mode.",
    ),
):
    """
    Install a workflow.
    """
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    if not nest.bin_dir_in_path():
        bin_dir_yellow = typer.style(nest.bin_dir, fg=typer.colors.YELLOW, bold=False)
        typer.echo(f"Please add SNK_BIN to your $PATH: {bin_dir_yellow}")
    if not Path(workflow).exists() and not workflow.startswith("http"):
        workflow = f"https://github.com/{workflow}.git"
    try:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task(description="Installing...", total=None)
            installed_workflow = nest.install(
                workflow,
                editable=editable,
                name=name,
                tag=tag,
                commit=commit,
                config=config,
                snakefile=snakefile,
                additional_resources=resource,
                force=force,
                conda=not no_conda,
                snakemake_version=snakemake_version,
                dependencies=dependencies,
                isolate=isolate,
            )
    except WorkflowExistsError as e:
        typer.secho(
            str(e) + ". Use a different name (--name) or overwrite (--force).",
            fg="red",
            err=True,
        )
        raise typer.Exit(1)
    except Exception as e:
        typer.secho(e, fg="red", err=True)
        raise typer.Exit(1)

    snk_config = SnkConfig.from_workflow_dir(installed_workflow.path, create_if_not_exists=True)
    version = snk_config.version
    version_str = f" ({version})" if version else ""
    typer.secho(f"Successfully installed {installed_workflow.name}{version_str}!", fg="green")


@app.command()
def uninstall(
    ctx: typer.Context,
    name: str = typer.Argument(..., help="Name of the workflow to uninstall."),
    force: Optional[bool] = typer.Option(
        False, "--force", "-f", help="Force uninstall without asking."
    ),
):
    """
    Uninstall a workflow.
    """
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    try:
        uninstalled = nest.uninstall(name, force=force)
    except WorkflowNotFoundError as e:
        typer.secho(e, fg="red")
        raise typer.Exit(1)
    if uninstalled:
        typer.secho(f"Successfully uninstalled {name}!", fg="green")


@app.command()
def list(
    ctx: typer.Context,
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Show the workflow paths."),
):
    """
    List the installed workflows.
    """
    from rich.console import Console
    from rich.table import Table

    table = Table("Workflow", "Version", show_header=True, show_lines=True)
    if verbose:
        table.add_column("Path")
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    try:
        workflows = nest.workflows
    except FileNotFoundError:
        workflows = []
    for workflow in workflows:
        snk_config = SnkConfig.from_workflow_dir(workflow.path, create_if_not_exists=True)
        if snk_config.version == "editable":
            version_str = "[green]editable[/green]"
        else:
            version_str = f"[blue]{snk_config.version}[/blue]"
        if verbose:
            table.add_row(
                workflow.name,
                version_str,
                f"[yellow]{str(workflow.path.resolve())}[/yellow]",
            )
        else:
            table.add_row(workflow.name, version_str)
    console = Console()
    console.print(table)

@app.command()
def create(path: Path, force: bool = typer.Option(False, "--force", "-f")):
    """Create a default snk.yaml project that can be installed with snk"""
    if path.exists():
        if not force:
            typer.secho(f"Directory '{path}' already exists! Use --force to overwrite.", fg="red", err=True)
            raise typer.Exit(1)
        else:
            typer.secho(f"Overwriting existing directory '{path}'.", fg="yellow", err=True)
            import shutil
            shutil.rmtree(path)
    try:
        path.mkdir(parents=True, exist_ok=False)
    except FileExistsError:
        typer.secho(f"Directory '{path}' already exists! Use --force to overwrite.", fg="red", err=True)
        raise typer.Exit(1)
    snk_config = SnkConfig.from_workflow_dir(path, create_if_not_exists=True)
    snk_config.cli["msg"] = {
        "help": "Print a help message.",
        "default": "Hello, World!",
        "required": False,
        "short": "m",
    }
    snk_config.save()
    typer.echo(f"Created snk.yaml at {snk_config._snk_config_path}")
    with open(path / "Snakefile", "w") as f:
        f.write("""rule all:\n    shell: f"echo {config['msg']}"\n""")

@app.command()
def edit(
    ctx: typer.Context,
    workflow_name: str = typer.Argument(
        ..., help="Name of the workflow to configure."
    ),
    path: bool = typer.Option(
        False, "--path", "-p", help="Show the path to the snk.yaml file."
    ),
):
    """
    Access the snk.yaml configuration file for a workflow.
    """
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    try:
        workflows = nest.workflows
    except FileNotFoundError:   
        workflows = []
    workflow = next((w for w in workflows if w.name == workflow_name), None) 
    if not workflow:
        typer.secho(f"Workflow '{workflow_name}' not found!", fg="red", err=True)
        raise typer.Exit(1)
    snk_config = SnkConfig.from_workflow_dir(workflow.path, create_if_not_exists=True)
    snk_config.save()
    if path:
        typer.echo(snk_config._snk_config_path)
    else:
        try:
            open_text_editor(snk_config._snk_config_path)
        except Exception as e:
            typer.secho(str(e), fg="red", err=True)
            raise typer.Exit(1)

# @app.command()
# def run(
#         workflow: str = typer.Argument(
#             ..., help="URL or Github name (user/repo) of the workflow to install."
#         ),
#     ):
#     """
#     Run the workflow in a temporary environment.
#     """
#     raise NotImplementedError

# @app.command()
# def annotations(config: Path):
#     """Generate annotations defaults from config file"""
#     raise NotImplementedError

# @app.command()
# def update():
#     """
#     Update a workflow.
#     """
#     raise NotImplementedError



from pathlib import Path

from snk import Nest


def test_init(bin_dir, snk_home):
    nest = Nest(snk_home, bin_dir=bin_dir)
    assert nest.snk_workflows_dir == Path(snk_home) / "workflows"
    for path in [nest.snk_workflows_dir, nest.bin_dir, nest.snk_home]:
        assert path.exists()


def test_download(nest: Nest):
    path = nest.download("https://github.com/Wytamma/snk-basic-pipeline.git", "rna-seq-star-deseq2")
    expected_location = nest.snk_workflows_dir / "rna-seq-star-deseq2"
    assert (expected_location).exists()
    assert path == expected_location


def test_create_package(nest: Nest):
    test_workflow_path = nest.snk_workflows_dir / "workflow-name"
    test_workflow_path.mkdir()
    path = nest.create_executable(workflow_path=test_workflow_path, name=test_workflow_path.name)
    assert path == nest.snk_home / "bin" / "workflow-name"


def test_install(nest: Nest):
    workflow = nest.install("tests/data/workflow")
    assert workflow.name == "workflow"
    assert len(nest.workflows) == 1
    assert (nest.snk_workflows_dir / "workflow").exists()
    assert (nest.snk_home / "bin" / "workflow").exists()

    with open(nest.snk_home / "bin" / "workflow") as f:
        contents = f.read()
        assert "from snk_cli import CLI" in contents
        assert "create_cli" in contents


def test_install_custom_name(nest: Nest):
    workflow = nest.install("tests/data/workflow", name="custom-name")
    assert workflow.name == "custom-name"


def test_link_workflow_executable_to_bin(nest: Nest):
    workflow_executable_path = Path("tests/data/bin/snk-basic-pipeline")
    executable_path = nest.link_workflow_executable_to_bin(workflow_executable_path)
    assert executable_path.exists() is True and executable_path.is_symlink() is True


def test_uninstall(nest: Nest):
    workflow = nest.install("tests/data/workflow")
    assert len(nest.workflows) == 1
    nest.uninstall(workflow.name, force=True)
    assert len(nest.workflows) == 0
    assert not (nest.snk_workflows_dir / "workflow").exists()
    assert not (nest.snk_home / "bin" / "workflow").exists()
